from __future__ import absolute_import

import uuid

from django.contrib import admin

from .models import Avatar, Game, User, Worksheet


class GameDataAdmin(admin.ModelAdmin):
    search_fields = ["id", "name", "owner__username", "owner__email"]
    list_display = ["id", "name", "owner"]
    raw_id_fields = ["owner", "main_user", "can_play", "game_class"]
    readonly_fields = ["players", "auth_token"]

    def players(self, obj):
        players = "\n".join(
            [student.new_user.first_name for student in obj.game_class.students.all()]
        )
        players = players.join(obj.game_class.teacher.new_user.first_name)
        return players


class AvatarDataAdmin(admin.ModelAdmin):
    search_fields = ["owner__username", "owner__email"]
    list_display = ["id", "owner_name", "game_name"]
    raw_id_fields = ["game"]
    readonly_fields = ["owner", "auth_token"]

    def owner_name(self, obj):
        return obj.owner

    def game_name(self, obj):
        return obj.game


class WorksheetDataAdmin(admin.ModelAdmin):
    search_fields = ["id", "name", "era"]
    list_display = ["id", "name", "era"]


admin.site.register(Game, GameDataAdmin)
admin.site.register(Avatar, AvatarDataAdmin)
admin.site.register(Worksheet, WorksheetDataAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-10-06 12:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0020_add_info_to_worksheet'),
    ]

    operations = [
        migrations.AddField(
            model_name='worksheet',
            name='student_pdf_name',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='worksheet',
            name='teacher_pdf_name',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



from dataclasses import dataclass
from random import randint

from simulation.cell import Cell
from simulation.worksheet import WORKSHEET, WorksheetData


@dataclass
class Obstacle:
    """A dataclass representing an obstacle in the simulation."""

    texture_choice: int
    orientation = "north"
    width: int = 1
    height: int = 1

    def serialize(self, cell: Cell):
        return {"location": cell.location.serialize(), "texture": self.texture_choice}

    def make_obstacle(worksheet: WorksheetData = WORKSHEET) -> "Obstacle":
        """
        Returns an obstacle with a randomly generated texture choice based on number of different
        obstacle textures indicated in the worksheet.

        Args:
            worksheet (WorksheetData, optional): The worksheet to use to generate the obstacle. Defaults to WORKSHEET.
        """
        texture_choice = randint(1, worksheet.number_of_obstacle_textures)
        return Obstacle(texture_choice)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-07-23 11:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0001_initial'),
        ('aimmo', '0015_game_worksheet'),
    ]

    operations = [
        migrations.AddField(
            model_name='game',
            name='game_class',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='games_for_class', to='common.Class'),
        ),
    ]



"""
Module for keeping track of inactivity for a given game.
"""

import asyncio
import logging
from enum import Enum

from requests import codes
from types import CoroutineType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simulation.django_communicator import DjangoCommunicator

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

SECONDS_TILL_CONSIDERED_INACTIVE = 3600


class StatusOptions(Enum):
    RUNNING = "r"
    PAUSED = "p"
    STOPPED = "s"


class ActivityMonitor:
    """
    Keeps track of the number of users currently connected.

    If no users are connected (measured by sockets) after a certain length
    of time, the game is marked as stopped and the pods will be shut down shortly after
    """

    def __init__(self, django_communicator: "DjangoCommunicator"):
        self.timer = Timer(
            SECONDS_TILL_CONSIDERED_INACTIVE, self.change_status_to_stopped
        )
        self.active_users = 0
        self.django_communicator = django_communicator

    def _start_timer(self):
        if self.timer.cancelled():
            LOGGER.info("No socket connections found. Timer started!")
            self.timer = Timer(
                SECONDS_TILL_CONSIDERED_INACTIVE, self.change_status_to_stopped
            )

    def _stop_timer(self):
        if not self.timer.cancelled():
            LOGGER.info("Cancelling timer!")
            self.timer.cancel()

    @property
    def active_users(self):
        return self.__active_users

    @active_users.setter
    def active_users(self, value: float):
        self.__active_users = value
        if self.__active_users:
            self._stop_timer()
        else:
            self._start_timer()

    async def change_status_to_stopped(self):
        LOGGER.info("Timer expired! Marking game as STOPPED")
        response = await self.django_communicator.patch_game(
            {"status": StatusOptions.STOPPED.value}
        )

        if response.status != codes["ok"]:
            LOGGER.error(f"Game could not be stopped. {response}")


class Timer:
    """
    Generic Timer with callback.

    This sleeps for `timeout=X` seconds, after X seconds the
    callback function is called, this happens asynchronously.
    """

    def __init__(self, timeout: float, callback: CoroutineType):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job())

    async def _job(self):
        await asyncio.sleep(self._timeout)
        await self._callback()

    def cancel(self):
        self._task.cancel()

    def cancelled(self):
        return self._task.cancelled()



from simulation.interactables.pickups.artefact import Artefact
from simulation.interactables.pickups.damage_boost_pickup import DamageBoostPickup
from simulation.interactables.pickups.health_pickup import HealthPickup
from simulation.interactables.pickups.invulnerability_pickup import (
    InvulnerabilityPickup,
)


def serialize_pickups(world_map):
    return [cell.interactable.serialize() for cell in world_map.pickup_cells()]


ALL_PICKUPS = (DamageBoostPickup, InvulnerabilityPickup, HealthPickup, Artefact)
__all__ = [
    "ALL_PICKUPS",
    "DamageBoostPickup",
    "InvulnerabilityPickup",
    "HealthPickup",
    "Artefact",
]



from common.models import Class
from django.core.exceptions import ValidationError
from django.db.models.query import QuerySet
from django.forms import ModelChoiceField, ModelForm, Select

from aimmo.models import Game, Worksheet


class AddGameForm(ModelForm):
    def __init__(self, classes: QuerySet, *args, **kwargs):
        super(AddGameForm, self).__init__(*args, **kwargs)
        self.fields["game_class"].queryset = classes
        self.fields["worksheet"].queryset = Worksheet.objects.all()

    game_class = ModelChoiceField(
        queryset=None, widget=Select, label="Class", required=True
    )
    worksheet = ModelChoiceField(
        queryset=None, widget=Select, label="Challenge", required=True
    )

    class Meta:
        model = Game
        exclude = [
            "Main",
            "name",
            "owner",
            "auth_token",
            "completed",
            "main_user",
            "static_data",
            "can_play",
            "public",
            "generator",
            "target_num_cells_per_avatar",
            "target_num_score_locations_per_avatar",
            "score_despawn_chance",
            "target_num_pickups_per_avatar",
            "pickup_spawn_chance",
            "obstacle_ratio",
            "start_height",
            "start_width",
            "status",
        ]

    def full_clean(self) -> None:
        super(AddGameForm, self).full_clean()
        try:
            self.instance.validate_unique()
        except ValidationError as e:
            self._update_errors(e)

    def clean(self):
        game_class: Class = self.cleaned_data.get("game_class")
        worksheet: Worksheet = self.cleaned_data.get("worksheet")

        if game_class and not Class.objects.filter(pk=game_class.id).exists():
            raise ValidationError("Sorry, an invalid class was entered")

        if worksheet and not Worksheet.objects.filter(pk=worksheet.id).exists():
            raise ValidationError("Sorry, an invalid challenge was entered")

        return self.cleaned_data



from simulation.interactables.conditions import avatar_on_cell, in_backpack
from simulation.interactables.effects import ArtefactEffect
from simulation.interactables.interactable import _Interactable


class Artefact(_Interactable):
    def __init__(self, cell):
        super(Artefact, self).__init__(cell)
        self.delete_after_effects_applied = True
        self.in_backpack = False
        self.conditions = [avatar_on_cell, in_backpack]
        self.effects.append(ArtefactEffect)

    def get_targets(self):
        return [self.cell.avatar]

    def __repr__(self):
        return "Artefact(Location={})".format(self.cell.location)

    def serialize(self):
        serialized_artefact = {"type": "artefact"}

        if not self.in_backpack:
            serialized_artefact["location"] = {
                "x": self.cell.location.x,
                "y": self.cell.location.y,
            }

        return serialized_artefact



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-09-09 14:26
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0001_initial'),
        ('aimmo', '0018_set_worksheet_2_as_default_for_games'),
    ]

    operations = [
        migrations.AlterField(
            model_name='game',
            name='game_class',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='games_for_class', to='common.Class', verbose_name='Class'),
        ),
        migrations.AlterField(
            model_name='game',
            name='worksheet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='aimmo.Worksheet'),
        ),
        migrations.AlterUniqueTogether(
            name='game',
            unique_together=set([('game_class', 'worksheet')]),
        ),
    ]



import math
from logging import getLogger

from simulation.cell import Cell
from simulation.game_logic import SpawnLocationFinder
from simulation.interactables.pickups import ALL_PICKUPS
from simulation.interactables.score_location import ScoreLocation
from simulation.level_settings import DEFAULT_LEVEL_SETTINGS
from simulation.location import Location
from typing import List

LOGGER = getLogger(__name__)


class WorldMap(object):
    """
    The non-player world state.
    """

    def __init__(self, grid, settings):
        """
        :param grid: All types of cells to be inserted into the map.
        :param settings: Constant values provided when generating a level/map.
        """
        self.grid = grid
        self.settings = settings
        self._spawn_location_finder = SpawnLocationFinder(self)

    @classmethod
    def _min_max_from_dimensions(cls, height, width):
        """
        The value provided by the user will be an integer both for the width and height
        components. We calculate the maximum and minimum dimensions in all directions.
        """
        max_x = int(math.floor(width / 2))
        min_x = -(width - max_x - 1)
        max_y = int(math.floor(height / 2))
        min_y = -(height - max_y - 1)
        return min_x, max_x, min_y, max_y

    @classmethod
    def generate_empty_map(cls, height, width, settings):
        new_settings = DEFAULT_LEVEL_SETTINGS.copy()
        new_settings.update(settings)

        (min_x, max_x, min_y, max_y) = WorldMap._min_max_from_dimensions(height, width)
        grid = {}
        for x in range(min_x, max_x + 1):
            for y in range(min_y, max_y + 1):
                location = Location(x, y)
                grid[location] = Cell(location)
        return cls(grid, new_settings)

    def all_cells(self) -> List[Cell]:
        return self.grid.values()

    def interactable_cells(self):
        return (cell for cell in self.all_cells() if cell.interactable)

    def score_cells(self):
        return (
            cell
            for cell in self.all_cells()
            if isinstance(cell.interactable, ScoreLocation)
        )

    def pickup_cells(self):
        return (
            cell
            for cell in self.all_cells()
            if isinstance(cell.interactable, ALL_PICKUPS)
        )

    def is_on_map(self, location):
        try:
            self.grid[location]
        except KeyError:
            return False
        return True

    def get_cell(self, location) -> Cell:
        try:
            return self.grid[location]
        except KeyError:
            # For backwards-compatibility, this throws ValueError
            raise ValueError("Location %s is not on the map" % location)

    def get_cell_by_coords(self, x, y):
        return self.get_cell(Location(x, y))

    def clear_cell_actions(self, location):
        try:
            cell = self.get_cell(location)
            cell.actions = []
        except ValueError:
            return

    def max_y(self):
        return max(self.grid.keys(), key=lambda c: c.y).y

    def min_y(self):
        return min(self.grid.keys(), key=lambda c: c.y).y

    def max_x(self):
        return max(self.grid.keys(), key=lambda c: c.x).x

    def min_x(self):
        return min(self.grid.keys(), key=lambda c: c.x).x

    @property
    def num_rows(self):
        return self.max_y() - self.min_y() + 1

    @property
    def num_cols(self):
        return self.max_x() - self.min_x() + 1

    @property
    def num_cells(self):
        return self.num_rows * self.num_cols

    def can_move_to(self, target_location):
        if not self.is_on_map(target_location):
            return False
        cell = self.get_cell(target_location)

        return (
            cell.habitable
            and (not cell.is_occupied or cell.avatar.is_moving)
            and len(cell.moves) <= 1
        )

    def attackable_avatar(self, target_location):
        """
        Return a boolean if the avatar is attackable at the given location (or will be
        after next move), else return None.
        """
        try:
            cell = self.get_cell(target_location)
        except ValueError:
            return None

        if cell.avatar:
            return cell.avatar

        if len(cell.moves) == 1:
            return cell.moves[0].avatar

        return None

    def get_no_fog_distance(self):
        return self.settings["NO_FOG_OF_WAR_DISTANCE"]

    def get_partial_fog_distance(self):
        return self.settings["PARTIAL_FOG_OF_WAR_DISTANCE"]

    def get_random_spawn_location(self):
        return self._spawn_location_finder.get_random_spawn_location()

    def __repr__(self):
        return repr(self.grid)

    def __iter__(self):
        return (
            (
                self.get_cell(Location(x, y))
                for y in range(self.min_y(), self.max_y() + 1)
            )
            for x in range(self.min_x(), self.max_x() + 1)
        )

    # Serialisation Utilities
    def get_serialized_south_west_corner(self):
        """
        Used in serialising the map size when sent to the front end. Very lightweight as
        it consists of two integers.

        :return: A dictionary with two values, x and y coordinates for the bottom left
        (south-west) corner of the map.
        """
        return {"x": self.min_x(), "y": self.min_y()}

    def get_serialized_north_east_corner(self):
        """
        Used in serialising the map size when sent to the front end. Very lightweight as
        it consists of two integers.

        :return: A dictionary with two values, x and y coordinates for the top right
        (north-west) corner of the map.
        """
        return {"x": self.max_x(), "y": self.max_y()}

    def serialize_score_location(self):
        """
        Used to serialize the score locations on every update.

        :return: A single list that contains all score locations. Within
        the list there are x and y coordinates.
        """

        def get_coords(cell):
            return {"location": {"x": cell.location.x, "y": cell.location.y}}

        return [
            get_coords(cell)
            for cell in self.all_cells()
            if isinstance(cell.interactable, ScoreLocation)
        ]

    def serialize_obstacles(self):
        """
        Used to serialize the obstacle locations on every update.

        :return: A list that contains all the obstacle information generated by inner method.
        """

        def serialize_obstacle(cell):
            return {
                "location": {"x": cell.location.x, "y": cell.location.y},
                "width": 1,
                "height": 1,
                "type": "wall",
                "orientation": "north",
            }

        return [
            cell.obstacle.serialize(cell)
            for cell in self.all_cells()
            if not cell.habitable
        ]


def WorldMapStaticSpawnDecorator(world_map, spawn_location):
    world_map._spawn_location_finder.get_random_spawn_location = lambda: spawn_location
    return world_map



import math
import random
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from logging import getLogger

from simulation.cell import Cell
from simulation.interactables.pickups import Artefact
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location

LOGGER = getLogger(__name__)

MapContext = namedtuple("MapContext", "num_avatars")


class _MapUpdater:
    __metaclass__ = ABCMeta

    @abstractmethod
    def update(self, world_map, context):
        raise NotImplementedError


class ScoreLocationUpdater(_MapUpdater):
    def update(self, world_map, context):
        for cell in world_map.score_cells():
            if random.random() < world_map.settings["SCORE_DESPAWN_CHANCE"]:
                cell.interactable.delete()

        new_num_score_locations = len(list(world_map.score_cells()))
        target_num_score_locations = int(
            math.ceil(
                context.num_avatars
                * world_map.settings["TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR"]
            )
        )
        num_score_locations_to_add = (
            target_num_score_locations - new_num_score_locations
        )
        locations = world_map._spawn_location_finder.get_random_spawn_locations(
            num_score_locations_to_add
        )
        for cell in locations:
            cell.interactable = ScoreLocation(cell)


class PickupUpdater(_MapUpdater):
    """
    Generates artefacts based on the TARGET_NUM_PICKUPS_PER_AVATAR setting.
    """

    def update(self, world_map, context):
        target_num_pickups = int(
            math.ceil(
                context.num_avatars
                * world_map.settings["TARGET_NUM_PICKUPS_PER_AVATAR"]
            )
        )
        max_num_pickups_to_add = target_num_pickups - len(
            list(world_map.pickup_cells())
        )
        locations = world_map._spawn_location_finder.get_random_spawn_locations(
            max_num_pickups_to_add
        )
        for cell in locations:
            cell.interactable = Artefact(cell)
            LOGGER.info("Adding new pickup at %s of type %s", cell, cell.interactable)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-11-13 13:34
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0021_add_pdf_names_to_worksheet'),
    ]

    operations = [
        migrations.AlterField(
            model_name='game',
            name='name',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



import json

from rest_framework import serializers

from aimmo.models import Game


class GameSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=100)
    settings = serializers.SerializerMethodField("get_settings_as_dict")
    status = serializers.CharField(max_length=1)
    class_id = serializers.SerializerMethodField()
    worksheet_id = serializers.SerializerMethodField()
    era = serializers.SerializerMethodField("get_worksheet_era")

    def get_class_id(self, game: Game):
        try:
            return str(game.game_class.id)
        except AttributeError:
            return None

    def get_worksheet_id(self, game: Game):
        try:
            return str(game.worksheet.id)
        except AttributeError:
            return "2"

    def get_worksheet_era(self, game: Game):
        try:
            return str(game.worksheet.era)
        except AttributeError:
            return "1"

    def get_settings_as_dict(self, game: Game):
        return json.dumps(game.settings_as_dict(), sort_keys=True)

    def update(self, instance, validated_data):
        instance.name = validated_data.get("name", instance.name)
        instance.status = validated_data.get("status", instance.status)
        instance.save()
        return instance



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-03-02 13:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('aimmo', '0011_reset_game_tokens'),
    ]

    operations = [
        migrations.AlterField(
            model_name='game',
            name='generator',
            field=models.CharField(
                choices=[('Main', 'Open World'), ('Level1', 'Level 1')], default='Main',
                max_length=20),
        ),
        migrations.AlterField(
            model_name='game',
            name='status',
            field=models.CharField(
                choices=[('r', 'running'), ('s', 'stopped'), ('p', 'paused')],
                default='r', max_length=1),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2020-06-29 10:17
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0013_alter_game_can_play'),
    ]

    operations = [
        migrations.CreateModel(
            name='Worksheet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('era', models.PositiveSmallIntegerField(choices=[(1, 'future'), (2, 'ancient'), (3, 'modern day'), (4, 'prehistoric'), (5, 'broken future')], default=1)),
                ('starter_code', models.TextField()),
            ],
        ),
    ]



import asyncio
import logging
import threading
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from simulation.action import PRIORITIES
from simulation.game_logic import EffectApplier, MapContext, PickupUpdater
from simulation.worksheet import WORKSHEET, WorksheetData

if TYPE_CHECKING:
    from turn_collector import CollectedTurnActions

LOGGER = logging.getLogger(__name__)

TURN_INTERVAL = 2


class SimulationRunner(object):
    """
    Game loop
    """

    daemon = True
    __metaclass__ = ABCMeta

    def __init__(self, game_state, communicator, worksheet: WorksheetData = WORKSHEET):
        self.game_state = game_state
        self.communicator = communicator
        self.worksheet: WorksheetData = WORKSHEET
        self._lock = threading.RLock()

    @abstractmethod
    async def run_turn(self, collected_turn_actions: "CollectedTurnActions"):
        pass

    async def _run_turn_for_avatar(self, avatar, serialized_action):
        """
        Send an avatar its view of the game state and register its
        chosen action & logs.
        """
        self._register_actions(avatar, serialized_action)

    def _register_actions(self, avatar, serialized_action):
        """
        Calls a function that constructs the action object, does error handling,
        and finally registers it onto the avatar.

        :param avatar: Avatar wrapper object
        :param serialized_action: A string representing the action
        """
        if avatar.decide_action(serialized_action):
            avatar.action.register(self.game_state.world_map)

    def _update_effects(self):
        with self._lock:
            for avatar in self.game_state.avatar_manager.active_avatars:
                avatar.update_effects()

    def update_environment(self):
        with self._lock:
            self._update_effects()
            num_avatars = len(self.game_state.avatar_manager.active_avatars)
            self.update(num_avatars, self.game_state)

    def update(self, num_avatars, game_state):
        EffectApplier().apply(game_state)
        self._update_map(num_avatars)

    def _update_map(self, num_avatars):
        context = MapContext(num_avatars=num_avatars)
        for MapUpdater in self.worksheet.map_updaters:
            MapUpdater().update(self.game_state.world_map, context=context)

    def add_avatar(self, player_id, location=None):
        with self._lock:
            location = (
                self.game_state.world_map.get_random_spawn_location()
                if location is None
                else location
            )
            avatar = self.game_state.avatar_manager.add_avatar(player_id, location)
            self.game_state.world_map.get_cell(location).avatar = avatar

    def remove_avatar(self, player_id):
        with self._lock:
            try:
                avatar = self.game_state.avatar_manager.get_avatar(player_id)
            except KeyError:
                return
            self.game_state.world_map.get_cell(avatar.location).avatar = None
            self.game_state.avatar_manager.remove_avatar(player_id)

    def add_avatars(self, player_ids):
        for player_id in player_ids:
            self.add_avatar(player_id)

    def delete_avatars(self, player_ids):
        for player_id in player_ids:
            self.remove_avatar(player_id)

    async def run_single_turn(self, collected_turn_actions: "CollectedTurnActions"):
        await self.run_turn(collected_turn_actions)
        self.update_environment()


class SequentialSimulationRunner(SimulationRunner):
    async def run_turn(self, collected_turn_actions: "CollectedTurnActions"):
        """
        Get and apply each avatar's action in turn.
        """
        avatars = self.game_state.avatar_manager.active_avatars

        for avatar in avatars:
            await self._run_turn_for_avatar(
                avatar, collected_turn_actions.get_action_for_player(avatar.player_id)
            )
            location_to_clear = avatar.action.target_location
            avatar.action.process(self.game_state.world_map)
            self.game_state.world_map.clear_cell_actions(location_to_clear)


class ConcurrentSimulationRunner(SimulationRunner):
    async def async_map(self, func, iterable_args):
        futures = [func(*arg) for arg in iterable_args]
        await asyncio.gather(*futures)

    async def run_turn(self, collected_turn_actions: "CollectedTurnActions"):
        """
        Concurrently get the intended actions from all avatars and register
        them on the world map. Then apply actions in order of priority.
        """

        avatars = self.game_state.avatar_manager.active_avatars
        args = [
            (avatar, collected_turn_actions.get_action_for_player(avatar.player_id))
            for avatar in avatars
        ]
        await self.async_map(self._run_turn_for_avatar, args)

        # Waits applied first, then attacks, then moves.
        avatars.sort(key=lambda a: PRIORITIES[type(a.action)])

        locations_to_clear = {
            a.action.target_location for a in avatars if a.action is not None
        }

        for action in (a.action for a in avatars if a.action is not None):
            action.process(self.game_state.world_map)

        for location in locations_to_clear:
            self.game_state.world_map.clear_cell_actions(location)



"""
Any helper functions used for game creation.
"""
from __future__ import absolute_import

import os

from aimmo.models import Avatar, Game


def create_game(main_user, form):
    """
    Creates a Game by:
    - saving the form
    - setting default values
    - adding users who can play the game
    - creating an avatar for the main user.

    :param main_user: The user who requested game creation, and is the game owner.
    :param form: The form used to submit the creation of the game.
    :param users_to_add_to_game: List of User objects who are able to play this game.
    :return: The initialised Game object.
    """
    game = form.save(commit=False)
    game.generator = "Main"
    game.owner = main_user
    game.main_user = main_user
    game.save()
    create_avatar_for_user(main_user, game.id)
    return game


def create_avatar_for_user(user, game_id):
    """
    Creates an Avatar object for a user. Sets the initial code to simple avatar code
    (unless specified otherwise).

    :param user: The user the Avatar is for.
    :param game_id: The id of the game in which the Avatar is created.
    :return: The initialised Avatar object.
    """
    game: Game = Game.objects.get(id=game_id)
    initial_code = game.worksheet.starter_code
    avatar = Avatar.objects.create(owner=user, code=initial_code, game_id=game_id)
    return avatar



# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open("aimmo/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="aimmo",
    packages=find_packages(exclude=["*.tests", "*.tests.*"]),
    package_dir={"aimmo": "aimmo"},
    include_package_data=True,
    install_requires=[
        "django >= 1.10.8, <= 1.11.24",
        "django-autoconfig >= 0.3.6, < 1.0.0",
        "django-js-reverse",
        "djangorestframework>=3.8.2, < 3.9.0",
        "eventlet",
        "requests",
        "hypothesis",
        "cfl-common",
    ],
    tests_require=["docker >= 3.5, < 3.6", "kubernetes == 5.0.0", "PyYAML == 4.2b1"],
    version=version,
    zip_safe=False,
)
